"""Python versions of Guava-like checks."""
from typing import Any, Iterable, Tuple, TypeVar, Union

# Disable naming convention warnings for type aliases
# pylint: disable=invalid-name
# Type annotation from TypeShed for classinfo argument  of isinstance and issubclass

_ClassInfo = Union[type, Tuple[Union[type, Tuple], ...]]


T = TypeVar("T")


def check_not_none(x: T, msg: str = None) -> T:
    """
    Raise an error if the given argument is None.

    This returns its input so you can do::
        self.x = check_not_none(x)

    check_arg would have the same result but is less clear to the reader
    """
    if x is None:
        if msg:
            raise ValueError(msg)
        else:
            raise ValueError()
    else:
        return x


def check_arg(result: Any, msg: str = None, msg_args: Tuple = None) -> None:
    if not result:
        if msg:
            raise ValueError(msg % (msg_args or ()))
        else:
            raise ValueError()


def check_state(result: Any, msg: str = None):
    if not result:
        if msg:
            raise AssertionError(msg)
        else:
            raise AssertionError()


def check_args_are_none(*args, msg: str = None):
    for arg in args:
        check_arg(arg is None, msg)


def check_args_not_none(*args, msg: str = None):
    for arg in args:
        check_arg(arg is not None, msg)


def check_isinstance(item: T, classinfo: _ClassInfo) -> T:
    if not isinstance(item, classinfo):
        raise TypeError(
            "Expected instance of type {!r} but got type {!r} for {!r}".format(
                classinfo, type(item), item
            )
        )
    return item


def check_opt_isinstance(item: T, classinfo: _ClassInfo) -> T:
    """
    Checks something is ether None or an instance of a given class.

    Raises a TypeError otherwise
    """
    if item and not isinstance(item, classinfo):
        raise TypeError(
            "Expected instance of type {!r} but got type {!r} for {!r}".format(
                classinfo, type(item), item
            )
        )
    return item


def check_all_isinstance(items: Iterable[Any], classinfo: _ClassInfo):
    for item in items:
        check_isinstance(item, classinfo)


def check_issubclass(item, classinfo: _ClassInfo):
    if not issubclass(item, classinfo):
        raise TypeError(
            "Expected subclass of type {!r} but got {!r}".format(classinfo, type(item))
        )
    return item


def check_in(item: Any, legal_values: Iterable[Any], item_name: str = None) -> None:
    if item not in legal_values:
        # dynamic import here to break circular dependency
        # performance impact is minimal since import only happens on precondition failure,
        # which will presumably crash the program
        import vistautils.misc_utils

        item_msg = " {!s} to be".format(item_name) if item_name else ""
        raise ValueError(
            "Expected{!s} one of {!s} but got {!s}".format(
                item_msg, vistautils.misc_utils.str_list_limited(legal_values, 10), item
            )
        )



import inspect
import logging
import re
import shutil
from pathlib import Path
from typing import (
    Any,
    List,
    Mapping,
    Match,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    Iterable,
)

import yaml
from attr import attrs
from immutablecollections import ImmutableDict

from vistautils.attrutils import attrib_opt_immutable
from vistautils.io_utils import CharSink, is_empty_directory
from vistautils.misc_utils import eval_in_context_of_modules
from vistautils.preconditions import check_arg, check_isinstance

_logger = logging.getLogger(__name__)  # pylint:disable=invalid-name


class ParameterError(Exception):
    pass


ParamType = TypeVar("ParamType")  # pylint:disable=invalid-name


@attrs(frozen=True, slots=True)
class Parameters:
    """
    Configuration parameters for a program.

    A `Parameters` object can be thought of as a hierarchical dictionary mapping parameter name
    strings to arbitrary values.  Hierarchical levels within a parameter name can be indicated
    via `.`; the levels are called namespaces. So `hello.world.foo` means "look up the 'foo'
    parameter within the namespace 'world' which is itself within the namespace 'hello'.

    The advantage of using this class over simple dictionary of dictionaries is that it provides
    type-safe accessors for parameters which can do validation and ensure certain conditions hold.
    This tends to push the discovery of errors earlier in program execution (or even before program
    execution).

    `Parameters` objects can be represented in multiple formats. The only loader currently
    implemented is `YAMLParametersLoader`.

    You can check if a lookup of a parameter would be successful using the `in` operator.
    """

    _data: ImmutableDict[str, Any] = attrib_opt_immutable(ImmutableDict)

    def __attrs_post_init__(self) -> None:
        for key in self._data:
            check_arg(
                "." not in key, "Parameter keys cannot contain namespace separator '.'"
            )

    @staticmethod
    def empty() -> "Parameters":
        """
        A `Parameters` with no parameter mappings.
        """
        return Parameters.from_mapping(ImmutableDict.empty())

    @staticmethod
    def from_mapping(mapping: Mapping) -> "Parameters":
        """
        Convert a dictionary of dictionaries into a `Parameter`s

        The top-level dictionary becomes the top-level namespace.  Each mapping-valued parameter
        becomes a namepace.
        """
        check_isinstance(mapping, Mapping)
        ret: ImmutableDict.Builder[str, Any] = ImmutableDict.builder()
        for (key, val) in mapping.items():
            if isinstance(val, Mapping):
                ret.put(key, Parameters.from_mapping(val))
            else:
                # this case will also be triggered if the value is already a parameters object
                ret.put(key, val)
        return Parameters(ret.build())

    def creatable_directory(self, param: str) -> Path:
        """
        Get a directory which can be written to.

        Interprets the string-valued parameter `param` as a directory path and creates it if
        it does not exist.  This allows writing files within this directory without needing to
        remember to create it or its parent directories.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        ret.mkdir(parents=True, exist_ok=True)
        return ret

    def creatable_empty_directory(self, param: str, *, delete=False) -> Path:
        """
        Get an empty directory which can be written to.

        Interprets the string-valued parameter `param` as a directory path and creates it if
        it does not exist.  If the directory already exists and is non-empty, behavior depends
        on the value of the `delete` argument.  If False (the default), an exception will be
        raised.  If True, the directory and its contents will be deleted.

        This allows writing files within this directory without needing to remember to create it
        or its parent directories.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        if ret.is_dir():
            if not is_empty_directory(ret):
                if delete:
                    shutil.rmtree(str(ret))
                    ret.mkdir(parents=True, exist_ok=True)
                else:
                    raise ParameterError(
                        "Expected an empty directory for parameters {!s},"
                        "but got non-empty path {!s}".format(param, ret)
                    )
            return ret
        elif ret.exists():
            raise ParameterError(
                "Expected an empty directory for parameters {!s},"
                "but got non-directory {!s}".format(param, ret)
            )
        else:
            return self.creatable_directory(param)

    def creatable_file(self, param: str) -> Path:
        """
        Get a file path which can be written to.

        Interprets the string-valued parameter `param` as a file path and creates its parent
        directory if it does not exist.  This allows writing to this file without needing to
        remember to create its parent directories.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        ret.parent.mkdir(parents=True, exist_ok=True)
        return ret

    def optional_creatable_file(self, param: str) -> Optional[Path]:
        """
        Get a file path which can be written to, if specified.

        Just like `creatable_file` but returns `None` if the parameter is absent.
        """
        if param in self:
            return self.creatable_file(param)
        else:
            return None

    def existing_file(self, param: str) -> Path:
        """
        Gets a path for an existing file.

        Interprets the string-valued parameter `param` as a file path. Throws a `ParameterError`
        if the path does not exist or is not a file.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        if ret.exists():
            if ret.is_file():
                return ret
            else:
                raise ParameterError(
                    "For parameter "
                    + param
                    + ", expected an existing file but got existing non-file "
                    + str(ret)
                )
        else:
            raise ParameterError(
                "For parameter "
                + param
                + ", expected an existing file but got non-existent "
                + str(ret)
            )

    def optional_existing_file(self, param: str) -> Optional[Path]:
        """
        Gets a path for an existing file, if specified.

        Interprets the string-valued parameter `param` as a file path. Returns `None`
        if no parmeter by that name is present.  Throws a `ParameterError`
        if the path does not exist.
        """
        if param in self:
            return self.existing_file(param)
        else:
            return None

    def existing_directory(self, param: str) -> Path:
        """
        Gets a path for an existing directory.

        Interprets the string-valued parameter `param` as a directory path. Throws a
        `ParameterError` if the path does not exist or is not a directory.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = Path.resolve(Path(self.string(param)))
        if ret.exists():
            if ret.is_dir():
                return ret
            else:
                raise ParameterError(
                    "For parameter "
                    + param
                    + ", expected an existing directory but got existing "
                    "non-directory " + str(ret)
                )
        else:
            raise ParameterError(
                "For parameter "
                + param
                + ", expected an existing directory but got non-existent "
                + str(ret)
            )

    def optional_existing_directory(self, param: str) -> Optional[Path]:
        """
        Gets a path for an existing directory, if specified.

        Interprets the string-valued parameter `param` as a directory path.
        If the parameter is not present, returns `None`.  Throws a
        `ParameterError` if the path does not exist.
        """
        if param in self:
            return self.existing_directory(param)
        else:
            return None

    def string(
        self, param_name: str, valid_options: Optional[Iterable[str]] = None
    ) -> str:
        """
        Gets a string-valued parameter.

        Throws a `ParameterError` if `param` is not a known parameter.
        """
        ret = self.get(param_name, str)
        if valid_options is not None and ret not in valid_options:
            raise ParameterError(
                f"The value {ret} for the parameter {param_name} is not one of the valid options "
                f"{tuple(valid_options)}"
            )
        return ret

    def __contains__(self, param_name: str) -> bool:
        return self._private_get(param_name, optional=True) is not None

    def namespace(self, name: str) -> "Parameters":
        """
        Get the namespace with the given name.
        """
        return self.get(name, Parameters)

    def integer(self, name: str) -> int:
        """
        Gets an integer parameters.
        """
        return self.get(name, int)

    def positive_integer(self, name: str) -> int:
        """
        Gets an parameter with a positive integer value.
        """
        ret = self.integer(name)
        if ret > 0:
            return ret
        else:
            raise ParameterError(
                "For parameter {!s}, expected a positive integer but got {!s}".format(
                    name, ret
                )
            )

    def optional_boolean(self, name: str) -> Optional[bool]:
        """
        Gets a boolean parameter if present.

        Avoid the temptation to do `params.optional_boolean('foo') or default_value`. If there is
        a default, prefer `optional_boolean_with_default`
        """
        return self.get_optional(name, bool)

    def optional_boolean_with_default(self, name: str, default_value: bool) -> bool:
        """
        Gets a boolean parameter if present; otherwise returns the provided default.
        """
        ret = self.optional_boolean(name)
        if ret is not None:
            return ret
        else:
            return default_value

    def optional_namespace(self, name: str) -> Optional["Parameters"]:
        """
        Get the namespace with the given name, if possible.

        If the namespace is not present or there is a non-namespace parameter of the given name,
        If the namespace is not present or there is a non-namespace parameter of the given name,
        `None` is returned.
        """
        ret = self.get_optional(name, object)
        if isinstance(ret, Parameters):
            return ret
        else:
            return None

    def optional_float(self, name: str) -> Optional[float]:
        """
        Gets a float parameter if present.

        Consider the idiom `params.optional_float('foo') or default_value`
        """
        return self.get_optional(name, float)

    def arbitrary_list(self, name: str) -> List:
        """
        Get a list with arbitrary structure.
        """
        return self.get(name, List)

    def optional_arbitrary_list(self, name: str) -> Optional[List]:
        """
        Get a list with arbitrary structure, if available
        """
        return self.get_optional(name, List)

    def optional_evaluate(
        self,
        name: str,
        expected_type: Type[ParamType],
        *,  # type: ignore
        namespace_param_name: str = "value",
        special_values: Mapping[str, str] = ImmutableDict.empty(),
    ) -> Optional[ParamType]:  # type: ignore
        """
        Get a parameter, if present, interpreting its value as Python code.

        Same as `evaluate` except returning `None` if the requested parameter is not present.
        """
        if name in self:
            return self.evaluate(
                name,
                expected_type,
                namespace_param_name=namespace_param_name,
                special_values=special_values,
            )
        else:
            return None

    # type ignored because ImmutableDict.empty() has type Dict[Any, Any]
    def evaluate(
        self,
        name: str,
        expected_type: Type[ParamType],
        *,  # type: ignore
        context: Optional[Mapping] = None,
        namespace_param_name: str = "value",
        special_values: Mapping[str, str] = ImmutableDict.empty(),
    ) -> ParamType:
        """
        Get a parameter, interpreting its value as Python code.

        The result of the code evaluation will be the parameter value. In addition to being a
        simple parameter, "name" can point to a namespace. In this case, the "value" parameter
        within the namespace is evaluated (unless some other parameter has been specified as
        `namespace_param_name`), but the list of modules given by the "import" parameter
        in that namespace (if present) are first imported.

        If the type of the result of the evaluation doesn't match `expected_type`, a
        `ParameterError` is raised.

        If `context` is specified, evaluation will happen in the context given. If you want
        evaluation to happen in the calling context, pass `locals()`.

        Sometimes is it convenient to provide shortcuts for common cases. These can be specified
        in a `special_values` map whose keys are the special case values and whose values are the
        strings of expressions to be evaluated.
        """

        def handle_special_values(val: str) -> str:
            return special_values.get(val, val)

        namespace = self.optional_namespace(name)
        try:
            if namespace:
                return eval_in_context_of_modules(
                    handle_special_values(namespace.string(namespace_param_name)),
                    context or locals(),
                    context_modules=namespace.optional_arbitrary_list("import") or [],
                    expected_type=expected_type,
                )
            else:
                return eval_in_context_of_modules(
                    handle_special_values(self.string(name)),
                    context or locals(),
                    context_modules=[],
                    expected_type=expected_type,
                )
        except Exception as e:
            raise ParameterError(
                "Error while evaluating parameter {!s}".format(name)
            ) from e

    # type ignored because ImmutableDict.empty() has type Dict[Any, Any]
    def object_from_parameters(  # type: ignore
        self,
        name: str,
        expected_type: Type[ParamType],
        *,
        context: Optional[Mapping] = None,
        creator_namepace_param_name: str = "value",
        special_creator_values: Mapping[str, str] = ImmutableDict.empty(),
        default_creator: Optional[Any] = None,
    ) -> ParamType:
        """
        Get an object of `expected_type`, initialized by the parameters in `name`.

        If `name` is a namespace, the `value` parameter within it is evaluated to get a "creator".
        If the result of the evaluation is a function, that function is itself the creator. If
        the result of the evaluation is a class, its `from_parameters` static method
        taking a single `Parameters` object will be used as the creator, if it exists. Otherwise
        its constructor will be used without parameters. The creator is then
        called with the namespace as its argument and the result is returned.  If the result does
        not match `expected_type` an exception will be raised. Do not include generic type arguments
        in `expected_type`.

        If `name` is a string, the same process is followed exception the string is evaluated
        directly to get the "creator" and an empty `Parameters` is passed.

        You can specify a different field within a namespace to be evaluated besides 'value' by
        specifying `creator_namespace_param_name`.

        You can specify a default cretor to be used if none is specified with `default_creator`.

        You may specify additional context within which evaluation should happen with `context`.
        If you want evaluation to happen in the calling context, set this to `locals()`.

        For the user's convenience, you may include a map of special values to expression strings.
        If the expression to be evaluated exactly matches any key of this map, the value from the
        map will be evaluated instead.

        If the namespace contains the field `import`, it will be treated as a comma-separated list
        of modules to be imported before evaluation.
        """
        if name in self:
            creator = self.evaluate(
                name,
                object,
                context=context,
                namespace_param_name=creator_namepace_param_name,
                special_values=special_creator_values,
            )
        elif default_creator:
            creator = default_creator
        else:
            raise ParameterError(
                "No creator class specified when creating an object from {!s}".format(
                    name
                )
            )

        params_to_pass = self.optional_namespace(name) or Parameters.empty()
        if inspect.isclass(creator):
            if hasattr(creator, "from_parameters"):
                ret = getattr(creator, "from_parameters")(params_to_pass)
            else:
                ret = creator()  # type: ignore
        elif callable(creator):
            ret = creator(params_to_pass)
        else:
            raise ParameterError(
                "Expected a class with from_parameters or a callable but got {!s}".format(
                    creator
                )
            )

        if isinstance(ret, expected_type):
            return ret
        else:
            raise ParameterError(
                "When instantiating using from_parameters, expected {!s} but"
                " got {!s}".format(expected_type, ret)
            )

    def get(self, param_name: str, param_type: Type[ParamType]) -> ParamType:
        """
        Get a parameter with type-safety.

        Gets the given parameter, throwing a `ParameterError` if it is not of the specified type.

        Throws a `ParameterError` if the parameter is unknown.
        """

        ret = self._private_get(param_name)
        if isinstance(ret, param_type):
            return ret
        else:
            raise ParameterError(
                "When looking up parameter '{!s}', expected a value of type {!s}, but got {!s} "
                "of type {!s}".format(param_name, param_type, ret, type(ret))
            )

    def get_optional(
        self, param_name: str, param_type: Type[ParamType]
    ) -> Optional[ParamType]:
        """
        Get a parameter with type-safety.

        Gets the given parameter, throwing a `ParameterError` if it is not of the
        specified type.

        If the parameter is unknown, returns `None`
        """
        ret = self._private_get(param_name, optional=True)
        if not ret or isinstance(ret, param_type):
            return ret
        else:
            raise ParameterError(
                "When looking up parameter '{!s}', expected a value of type {!s}, but got {!s} "
                "of type {!s}".format(param_name, param_type, ret, type(ret))
            )

    def path_list_from_file(self, param: str, *, log_name=None) -> Sequence[Path]:
        """
        Gets a list of paths from the file pointed to by param

        The paths are assumed to be listed one-per-line. Blank lines and lines
        where the first non-whitespace character is '#' are skipped.

        If log_name is specified, a message will be logged at info level of the form "Loaded
        <number> <log_name> from <file>"
        """
        file_list_file = self.existing_file(param)
        with open(file_list_file, "r") as inp:
            ret = [
                Path(line.strip())
                for line in inp
                if line.strip() and not line.strip().startswith("#")
            ]
            if log_name:
                _logger.info("Loaded %s %s from %s", len(ret), log_name, file_list_file)
            return ret

    def path_map_from_file(self, param: str, *, log_name=None) -> Mapping[str, Path]:
        """
        Gets a map of keys to paths from the file pointed to by param

        We assume there are two tab-separated fields.  The first is the string key,
        the second is the path.

        If log_name is specified, a message will be logged at info level of the form "Loaded
        <number> <log_name> from <file>"
        """
        file_map_file = self.existing_file(param)
        with open(file_map_file, "r") as inp:
            ret_b: ImmutableDict.Builder[str, Path] = ImmutableDict.builder()
            for (line_num, line) in enumerate(inp):
                try:
                    parts = line.split("\t")
                    if len(parts) != 2:
                        raise IOError(
                            "Expected two tab-separated fields but got {!s}".format(
                                len(parts)
                            )
                        )
                    ret_b.put(parts[0].strip(), Path(parts[1].strip()))
                except Exception as e:
                    raise IOError(
                        "Error parsing line {!s} of {!s}:\n{!s}".format(
                            line_num, file_map_file, line
                        )
                    ) from e

            ret = ret_b.build()
            if log_name:
                _logger.info("Loaded %s %s from %s", len(ret), log_name, file_map_file)
            return ret

    def _private_get(self, param_name: str, optional=False) -> Any:
        # pylint:disable=protected-access
        param_components = param_name.split(".")
        check_arg(param_components, "Parameter name cannot be empty")

        current = self
        namespaces_processed = []
        for param_component in param_components:
            if not isinstance(current, Parameters):
                if optional:
                    return None
                else:
                    raise ParameterError(
                        "When getting parameter "
                        + param_name
                        + " expected "
                        + ".".join(namespaces_processed)
                        + " to be a map, but it is a leaf: "
                        + str(current)
                        + ". Maybe you mistakenly prefixed the map keys with '-'?"
                    )

            if param_component in current._data:
                current = current._data[param_component]
                namespaces_processed.append(param_component)
            elif not optional:
                if namespaces_processed:
                    context_string = "in context " + ".".join(namespaces_processed)
                else:
                    context_string = "in root context"
                available_parameters = str(
                    [
                        key
                        for (key, val) in current._data.items()
                        if not isinstance(val, Parameters)
                    ]
                )
                available_namespaces = str(
                    [
                        key
                        for (key, val) in current._data.items()
                        if isinstance(val, Parameters)
                    ]
                )
                raise ParameterError(
                    "Parameter "
                    + param_name
                    + " not found. In "
                    + context_string
                    + " available parameters are "
                    + available_parameters
                    + ", available namespaces are "
                    + available_namespaces
                )
            else:
                # absent optional parameter
                return None

        return current


class YAMLParametersLoader:
    """
    Loads `Parameters` from a modified YAML format.

    The format of the parameters file is YAML with the following restrictions:
    * all non-leaf objects, including the top-level, must be maps
    * all keys must be strings
    * duplicate keys are not allowed (this is not currently enforced: #259)

    The following additional processing will be done:
    * if there is a top-level list-valued key '_includes', it will be removed.  Each of its
    values will be interpreted as a path and loaded.  If there are multiple included files,
    parameters loaded from earlier files will form the loading context (See `load`) for later ones.
    Parameter mappings in files loaded later will overwrite those in files loaded earlier.  Relative
    paths will be interpreted relative to the parameter file being loaded.

    * any parameters containing strings surrounded by `%` will be interpolated by replacing that
    string (and the `%`s) with the value of the parameter.  Failed interpolations will result
    in a `ParameterError`.  You cannot currently interpolate from parameters in the same file,
    but this capability will be added in the future.

    See unit tests in `test_parameters` for examples.
    """

    def load(
        self, f: Union[str, Path], context=Parameters.empty(), root_path: Path = None
    ):
        """
        Loads parameters from a YAML file.

        If `context` is specified, its content will be included in the returned Parameters (if
        not overridden) and will be available for interpolation.

        If `root_path` is specified, that path is used for resolving relative path names in
        includes instead of the path of the parameter file being loaded.
        """
        if isinstance(f, str):
            f = Path(f)
        try:
            # if no special path is specified, included files will be resolved relative to
            # this file's path
            if not root_path:
                root_path = f.parent

            with open(f, "r") as ymlfile:
                raw_yaml = yaml.load(ymlfile)
                self._validate(raw_yaml)
            cur_context = context

            # process and remove special include directives
            if "_includes" in raw_yaml:
                for included_file in raw_yaml["_includes"]:
                    _logger.info("Processing included parameter file %s", included_file)
                    included_file_path = Path(root_path, included_file).resolve()
                    cur_context = self._unify(
                        cur_context,
                        self.load(
                            included_file_path, root_path=root_path, context=cur_context
                        ),
                    )
                del raw_yaml["_includes"]

            return self._unify(
                cur_context,
                self._interpolate(Parameters.from_mapping(raw_yaml), cur_context),
            )
        except Exception as e:
            raise IOError("Failure while loading parameter file " + str(f)) from e

    @staticmethod
    def _validate(raw_yaml: Mapping):
        # we don't use check_isinstance so we can have a custom error message
        check_arg(
            isinstance(raw_yaml, Mapping),
            "Parameters YAML files must be mappings at the top level",
        )
        YAMLParametersLoader._check_all_keys_strings(raw_yaml)

    @staticmethod
    def _check_all_keys_strings(mapping: Mapping, path=None):
        if path is None:
            path = []

        non_string_keys = [x for x in mapping.keys() if not isinstance(x, str)]
        if non_string_keys:
            context_string = (
                (" in context" + ".".join(path)) if path else " in root context"
            )
            raise IOError("Non-string key(s) " + str(non_string_keys) + context_string)

        for val in mapping.values():
            if isinstance(val, Mapping):
                YAMLParametersLoader._check_all_keys_strings(val)

    _INTERPOLATION_REGEX = re.compile("%((\\w|\\.)+)%")

    @staticmethod
    def _interpolate(to_interpolate: Parameters, context: Parameters) -> Parameters:
        # pylint:disable=protected-access
        def interpolate(key, raw_value) -> Any:
            def lookup_interpolation(interpolation_key: Match) -> str:
                try:
                    # we know it is safe to strip off the edge characters because they must
                    # be the wrapping '%'s
                    return context.string(interpolation_key.group()[1:-1])
                except ParameterError:
                    raise ParameterError(
                        "Exception while interpolating parameter "
                        + key
                        + " with raw value "
                        + raw_value
                    )

            if isinstance(raw_value, str):
                return YAMLParametersLoader._INTERPOLATION_REGEX.sub(
                    lookup_interpolation, raw_value
                )
            elif isinstance(raw_value, Parameters):
                # TODO: need topological sort. Issue #258
                return YAMLParametersLoader._interpolate(raw_value, context)
            else:
                return raw_value

        return Parameters.from_mapping(
            {key: interpolate(key, val) for (key, val) in to_interpolate._data.items()}
        )

    def _unify(self, old: Parameters, new: Parameters, namespace="") -> Parameters:
        #  pylint:disable=protected-access
        ret = dict()
        for (key, old_val) in old._data.items():
            if key in new:
                new_val = new._data[key]
                if isinstance(old_val, Parameters) != isinstance(new_val, Parameters):
                    raise IOError(
                        "When unifying parameters, "
                        + namespace
                        + key
                        + "is a parameter on one side and a namespace on the other"
                    )
                elif isinstance(old_val, Parameters):
                    ret[key] = self._unify(old_val, new_val, namespace + key + ".")
                else:
                    ret[key] = new_val
            else:
                ret[key] = old_val

        for (key, new_val) in new._data.items():
            if key not in old:
                ret[key] = new_val

        return Parameters.from_mapping(ret)


@attrs(frozen=True)
class YAMLParametersWriter:
    def write(self, params: Parameters, sink: Union[Path, str, CharSink]) -> None:
        # pylint:disable=protected-access

        def dictify(data):
            if isinstance(data, ImmutableDict):
                return {k: dictify(v) for (k, v) in data.items()}
            elif isinstance(data, Parameters):
                return dictify(data._data)
            else:
                return data

        if isinstance(sink, Path) or isinstance(sink, str):
            sink = CharSink.to_file(sink)
        with sink.open() as out:
            yaml.dump(
                dictify(params._data),
                out,
                # prevents leaf dictionaries from being written in the
                # human unfriendly compact style
                default_flow_style=False,
                indent=4,
                width=78,
            )



#!/usr/bin/env python

"""
Joins multiple character key-value stores into one.

This is useful when recombining the results of a corpus which has been split for parallel
processing.

Currently the input stores are always assumed to be standard zip file stores. This might become
configurable in the future.

The "input_store_list_file" parameter points to a file containing the paths of the stores to join.
The "output" namespace describes the key-value store to output to. See
char_key_value_sink_from_params for details.
"""
import sys
from typing import Set

from vistautils.parameters import Parameters, YAMLParametersLoader
from vistautils.key_value import KeyValueSource, char_key_value_sink_from_params


def main(params: Parameters):
    input_paths = params.path_list_from_file('input_store_list_file',
                                             log_name="input key-value stores")
    keys_written: Set[str] = set()
    with char_key_value_sink_from_params('output', params,
                                         eval_context=locals()) as out:
        for input_path in input_paths:
            with KeyValueSource.zip_bytes_source(input_path) as inp:
                for key in inp.keys():
                    if key in keys_written:
                        raise RuntimeError("Duplicate key: {}".format(key))
                    keys_written.add(key)
                    out[key] = inp[key]


if __name__ == '__main__':
    main(YAMLParametersLoader().load(sys.argv[1]))



from typing import Iterable, Optional, Sized, Tuple, TypeVar
from typing_extensions import Protocol

from attr import attrs, attrib
from immutablecollections import (
    ImmutableSet,
    ImmutableSetMultiDict,
    immutablesetmultidict,
)

from vistautils.attrutils import attrib_instance_of
from vistautils.preconditions import check_arg
from vistautils.range import Range


@attrs(frozen=True, slots=True, repr=False)  # pylint:disable=inherit-non-class
# Pylint disable due to https://github.com/PyCQA/pylint/issues/2472
class Span(Sized):
    """
    A range of character offsets.

    Inclusive of `start`, exclusive of `end`.

    You can test sub-span containment with the `in` operator: `Span(1,5) in Span(0, 6)`.
    For checking whether an offset lies in a span, use `contains_offset`
    """

    start: int = attrib_instance_of(int)
    end: int = attrib_instance_of(int)

    # noinspection PyUnusedLocal
    @start.validator
    def _validate_start(self, attr, val):  # pylint:disable=unused-argument
        check_arg(
            self.start < self.end,
            "Start offset must be strictly less then end offset but " "got [%s,%s)",
            (self.start, self.end),
        )

    @staticmethod
    def from_inclusive_to_exclusive(start_inclusive: int, end_exclusive: int) -> "Span":
        """
        Same as the constructor.

        But the more explicit name increases readability and reduces off-by-one errors.
        """
        return Span(start_inclusive, end_exclusive)

    def contains_offset(self, i: int) -> bool:
        return self.start <= i < self.end

    def contains_span(self, other: "Span") -> bool:
        return self.start <= other.start and other.end <= self.end

    def precedes(self, other: "Span") -> bool:
        """
        Get whether this span precedes another.

        To be judged as preceding, this span must end before other begins.
        """
        return self.end <= other.start

    def follows(self, other: "Span") -> bool:
        """
        Get whether this span follows another.

        To be judged as following, this span must start after the other ends.
        """
        return self.start >= other.end

    def overlaps(self, other: "Span") -> bool:
        """
        Get whether this span overlaps another.

        Two spans overlap if at least one offset position is in
        both of them.
        """
        # if they don't overlap, one must precede the other
        return not (self.precedes(other) or other.precedes(self))

    def as_range(self) -> Range[int]:
        return Range.closed(self.start, self.end - 1)

    def __contains__(self, item: "Span") -> bool:
        return self.contains_span(item)

    def __len__(self) -> int:
        return self.end - self.start

    def clip_to(self, enclosing: "Span") -> Optional["Span"]:
        """
        Get a copy of this span clipped to be entirely enclosed by another span.

        If this span lies entirely outside `enclosing`, then
        `None` is returned.
        """
        if not enclosing.overlaps(self):
            return None
        if enclosing.contains_span(self):
            return self
        return Span(max(self.start, enclosing.start), min(self.end, enclosing.end))

    def shift(self, shift_amount: int) -> "Span":
        """
        Get a copy of this span with both endpoints shifted.

        Negative values shift the span to the left, positive to the right.
        """
        return Span(self.start + shift_amount, self.end + shift_amount)

    @staticmethod
    def minimal_enclosing_span(spans: Iterable["Span"]) -> "Span":
        """
        Get the minimal span enclosing all given spans.

        This will raise a `ValueError` if `spans` is empty.
        """
        return Span(min(span.start for span in spans), max(span.end for span in spans))

    @staticmethod
    def earliest_then_longest_first_key(x: "Span") -> Tuple[int, int]:
        length = x.end - x.start
        return x.start, -length

    def __repr__(self):
        return "[%s:%s)" % (self.start, self.end)


class HasSpan(Protocol):
    __slots__: tuple = ()
    span: Span

    @property
    def start(self) -> int:
        return self.span.start

    @property
    def end(self) -> int:
        return self.span.end

    def contains_offset(self, i) -> bool:
        return self.span.contains_offset(i)

    def contains_span(self, other) -> bool:
        return self.span.contains_span(other)


T = TypeVar("T", bound=HasSpan)


class HasSpanIndex(Protocol[T]):
    """
    Support efficient lookup of items with spans.
    """

    def exactly_matching(self, span: Span) -> ImmutableSet[T]:
        """
        Gets all contained items whose spans match *span* exactly.
        """

    @staticmethod
    def index(items: Iterable[T]) -> "HasSpanIndex[T]":
        """
        Creates a ``HasSpanIndex`` for the given items.
        """
        return _OverLappingHasSpanIndex(
            # mypy is confused
            immutablesetmultidict(((item.span, item) for item in items))  # type: ignore
        )


@attrs(frozen=True, slots=True)
class _OverLappingHasSpanIndex(HasSpanIndex[T]):
    """
    An implementation of ``HasSpanIndex`` for items whose spans may overlap.
    """

    _span_to_item_index: ImmutableSetMultiDict[Span, T] = attrib(
        converter=immutablesetmultidict, default=immutablesetmultidict()  # type: ignore
    )

    def exactly_matching(self, span: Span) -> ImmutableSet[T]:
        return self._span_to_item_index[span]



import argparse
import os
import tarfile
from zipfile import ZipFile

USAGE = """
    Converts a (un)compressed tar file to .zip
    This is useful because .zip allows random access, but the LDC often distributes things as .tgz.
    zip version will be input name with .tar/.tar.gz/.tgz stripped (if present), .zip added
"""


def getargs():
    """Get command-line arguments."""
    parser = argparse.ArgumentParser(usage=USAGE)
    arg = parser.add_argument

    arg("tar_file_name",
        help="The input .tar/.tar.gz/.tgz file to be converted")

    arg("--dont-stream",
        action="store_true",
        help="Should the tar contents be added to the zip file by unpacking "
        "them to disk first, rather than streaming them?")

    arg("--omit-large-files",
        action="store_true",
        help="Should large files (larger than LARGE_FILE_CUTOFF) not be "
        "transferred from the tar file to the zip file?")

    arg("--large-file-cutoff",
        type=float,
        default=2,
        help="Files larger than this cutoff (in gigabytes) will not be added"
        "to the zip file. The default is 2 gigabytes.")

    return parser.parse_args()


def main():
    args = getargs()
    tar_file = args.tar_file_name
    output_zip_name = _output_name(tar_file)
    dont_stream = args.dont_stream
    omit_large_files = args.omit_large_files
    large_file_cutoff = args.large_file_cutoff * 1e9

    print(f"Copying tarball {tar_file} to {output_zip_name}")
    print("WARNING: This will throw away all metadata. It also might not work "
          "if you have files with /s in the name or other odd things. "
          "Please sanity check your output")

    if os.path.exists(output_zip_name):
        print(f"WARNING: Removing existing file at {output_zip_name}")
        os.remove(output_zip_name)

    with ZipFile(output_zip_name, 'x') as out:
        with tarfile.open(tar_file) as inp:
            if dont_stream:
                inp.extractall()
                for member in inp:
                    if member.isfile():
                        out.write(member.name)
                    os.remove(member.name)
            else:
                for member in inp:
                    if member.isfile():
                        if omit_large_files and member.size > large_file_cutoff:
                            print(f"WARNING: Omitting {member.name} as "
                                  f"its size {member.size / 1e9:.2f} GB "
                                  f"exceeds {args.large_file_cutoff:.2f} GB")
                        else:
                            print(f"Copying {member.name}")
                            with inp.extractfile(member) as data:
                                out.writestr(member.name, data.read())


def _output_name(filename: str) -> str:
    valid_extensions = [".tar", ".tgz", "tar.gz"]
    for valid_extension in valid_extensions:
        if filename.endswith(valid_extension):
            filename = filename[:len(filename) - len(valid_extension)]
            break

    return filename + ".zip"


if __name__ == '__main__':
    main()



from typing import TypeVar, Iterable, Tuple

import vistautils.iter_utils

# deprecated, present for backwards compatibility
_T = TypeVar("_T")


# noinspection PyTypeHints
def tile_with_pairs(iterable: Iterable[_T]) -> Iterable[Tuple[_T, _T]]:
    # noinspection PyTypeChecker
    return vistautils.iter_utils.windowed(  # type: ignore
        iterable, 2, partial_windows=False
    )



import importlib
from itertools import chain
from pathlib import Path
from typing import Any, Generic, Iterable, List, Mapping, Sequence, Type, TypeVar, Union

from attr import attrib, attrs

from vistautils import attrutils, preconditions


def str_list_limited(_list: Iterable[Any], limit: int) -> str:
    """
    Get the string representation of a list which is not too long.

    If the size of the input list is less or equal to the limit, returns str(list).
    If the size exceeds the limit, returns a string representation which matches str(list) for
    the first limit items, and the has the string "and N more", where N is the number of items
    not shown.

    This is useful for producing exception messages for collection lookups, so you can give
    the user some idea what is in the collection without potentially blowing up their terminal
    for large collections.
    """
    if limit < 0:
        raise ValueError("Limit must be positive")
    if not isinstance(_list, List):
        _list = list(_list)

    if len(_list) <= limit:  # type: ignore
        return str(_list)
    else:
        return (
            "["
            + ", ".join(repr(x) for x in _list[:limit])
            + " and %s more" % (len(_list) - limit)
            + "]"
        )


T = TypeVar("T")


def eval_in_context_of_modules(
    to_eval: str,
    context: Mapping[Any, Any],
    *,
    context_modules: Sequence[str],
    expected_type: Type[T]
) -> T:
    """
    Evaluate the given expression in the specified context.

    Optionally, enforce that the result is of the expected type, throwing a
    `RuntimeException` otherwise.

    The context of evaluation will be that given by `context` augmented by the import of the
    modules whose names are given in `context_modules`.  If you want this to be evaluated in the
    context of the call site, pass `locals` as the context.

    Just like with `eval` itself, never pass anything from an uncontrolled source to this method,
    since it could allow arbitrary code execution.
    """
    # we make a copy so we do not alter the calling context
    context = dict(context)

    # import into the context to be used for evaluation any additional modules requested
    for module_name in context_modules:
        package_parts = module_name.split(".")
        # emulate the import statement's behavior of importing parent packages
        for package_part_idx in range(len(package_parts)):
            package_name = ".".join(package_parts[0 : package_part_idx + 1])
            if package_name not in context:
                context[package_name] = importlib.import_module(package_name)
    ret = eval(to_eval, context)  # pylint:disable=eval-used
    if isinstance(ret, expected_type):
        return ret
    else:
        raise TypeError(
            "Expected result of evaluating {!s} to be of type {!s} but "
            "got {!s}".format(to_eval, expected_type, ret)
        )


def pathify(p: Union[str, Path]) -> Path:
    """
    Allow functions to take strings or proper `Path`s

    If the input is a `Path`, it is returned unchanged. If a string, it is changed to a `Path`
    """
    if isinstance(p, Path):
        return p
    else:
        return Path(p)


def strip_extension(name: str) -> str:
    """
    Remove a single extension from a file name, if present.
    """
    last_dot = name.rfind(".")
    if last_dot > -1:
        return name[:last_dot]
    else:
        return name


def flatten_once_to_list(iterable_of_iterables: Iterable[Iterable[T]]) -> List[T]:
    """
    Removes one level of nesting from nested iterables.

    Taken from the itertools recipes.
    """
    return list(chain.from_iterable(iterable_of_iterables))


# can't set slots=True or you get https://github.com/python-attrs/attrs/issues/313
@attrs(frozen=True)
class WithId(Generic[T]):
    """
    Pair some object with some ID.

    This is typically used as an input to an Ingester when no document ID can be extracted
    from the data itself.
    """

    # can't use attr_instance_of due to circular import problems
    id: str = attrib()
    item: T = attrib()

    def __attrs_post_init__(self) -> None:
        preconditions.check_arg(self.item is not None)
        preconditions.check_arg(isinstance(self.id, str), "Id must be a string")
        preconditions.check_arg(self.id, "Doc IDs may not be empty")


@attrs(frozen=True)
class Scored(Generic[T]):
    """
    An item together with a score.
    """

    item: T = attrib()
    score: float = attrutils.attrib_instance_of(float)

    def __attrs_post_init__(self) -> None:
        preconditions.check_arg(self.item is not None, "Item of a scored may not be None")
        preconditions.check_arg(isinstance(self.score, float))



"""
Code for applying HTML-like markup to text.

This is useful for rendering NLP annotations to users.

Adapted from ``edu.isi.nlp.strings.formatting`` from https://github.com/isi-vista/nlp-util, which
is itself derived from code from BBN Technologies.
"""
import io

import itertools
from typing import Optional, List, Mapping, Collection, Iterable

from attr import attrs, evolve
from immutablecollections import ImmutableDict

from vistautils.span import Span
from vistautils.attrutils import attrib_instance_of, attrib_opt_immutable
from vistautils.preconditions import check_arg

DIV = "div"
SPAN = "span"


@attrs(frozen=True, slots=True)
class AnnotatedSpan:
    """
    An HTML-like annotation applied to a span of offsets.

    The label is the primary label to be applied to the region.
    Additionally, key-value metadata (attributes) can be applied.

    When rendered as HTML, the primary label will become the tag and the metadata will
    becomes attributes.
    """

    label: str = attrib_instance_of(str)
    span: Span = attrib_instance_of(Span)
    attributes: Mapping[str, str] = attrib_opt_immutable(ImmutableDict)

    @staticmethod
    def create_div_of_class(span: Span, clazz: str) -> "AnnotatedSpan":
        return AnnotatedSpan(DIV, span, {"class": clazz})

    @staticmethod
    def create_span_of_class(span: Span, clazz: str) -> "AnnotatedSpan":
        return AnnotatedSpan(SPAN, span, {"class": clazz})


def to_start_tag(annotated_range: AnnotatedSpan) -> str:
    """
    Make the start tag of an HTML element from an `AnnotatedSpan`
    """
    key_value_string = " ".join(
        f'{k}="{v}"' for (k, v) in sorted(annotated_range.attributes.items())
    )
    attributes_string = (" " + key_value_string) if annotated_range.attributes else ""
    return f"<{annotated_range.label}{attributes_string}>"


def to_end_tag(annotated_range: AnnotatedSpan) -> str:
    """
    Make the end tag of an HTML element from an `AnnotatedOffsetRange`
    """
    return f"</{annotated_range.label}>"


@attrs(frozen=True, slots=True)
class HTMLStyleAnnotationFormatter:
    def annotated_text(
        self,
        text: str,
        annotations: Collection[AnnotatedSpan],
        *,
        text_offsets: Optional[Span] = None,
    ) -> str:
        """
        Mark annotations on text in an HTML-like style.

        Each annotation will becomes an HTML tag wrapping the text at the corresponding offsets.
        Any attributes will become HTML attributes.

        This does not add any other HTML annotations (`head`, `body`, etc.), so if desired the
        user should add them afterwards.

        If `text_offsets` is specified, the annotations are assumed to have offsets with respect
        to some larger string, where `text` is a substring of that string with offsets
        `text_offsets` relative to it.  You might use this, for example, to render a single
        paragraph from a document.
        """
        if not text_offsets:
            text_offsets = Span.from_inclusive_to_exclusive(0, len(text))
        check_arg(
            len(text_offsets) == len(text),
            f"Text offsets length {len(text_offsets)} "
            f"does not match text length {len(text)}",
        )

        # we process the annotations to (a) ensure they all fit within the requested snippet
        # and (b) shift their offsets so that all offsets are relative to the text being
        # formatted
        processed_annotations = self._clip_to_offsets_and_shift(annotations, text_offsets)

        ret = io.StringIO()
        last_uncopied_offset = 0
        for tag in self._tag_sequence(processed_annotations):
            if last_uncopied_offset < tag.offset:
                ret.write(text[last_uncopied_offset : tag.offset])
                last_uncopied_offset = tag.offset

            ret.write(tag.string)

        # get any trailing text after last tag
        if last_uncopied_offset < text_offsets.end:
            ret.write(text[last_uncopied_offset : text_offsets.end])
        return ret.getvalue()

    @staticmethod
    def _clip_to_offsets_and_shift(
        unclipped_annotations: Collection[AnnotatedSpan], text_offsets: Span
    ) -> List[AnnotatedSpan]:
        """
        Clip or filter out annotations to ensure all are within the text being formatted.
        """
        # there are three cases:
        #    (a) an annotation is entirely contained in the text range being formatted and
        #              doesn't need alteration.
        #    (b) an annotation is partly contained in the text range being formatted and needs
        #              to be trimmed to git.
        #    (c) an annotation lies entirely outside the text range being formatted and should
        #              be omitted
        ret: List[AnnotatedSpan] = []
        for unclipped_annotation in unclipped_annotations:
            clipped_annotation_span = unclipped_annotation.span.clip_to(text_offsets)
            if clipped_annotation_span:
                # this is both cases a and b - in case (a) `clip_to` above returns the span
                # unchanged
                # we now need to shift the annotation offsets so that they are relative to the
                # snippet being formatted
                shifted_annotation_span = clipped_annotation_span.shift(
                    -text_offsets.start
                )
                ret.append(evolve(unclipped_annotation, span=shifted_annotation_span))
            # otherwise, we are in case (c) and we drop the annotation
        return ret

    @attrs(frozen=True, slots=True)
    class Tag:
        string: str = attrib_instance_of(str)
        is_start: bool = attrib_instance_of(bool)
        offset: int = attrib_instance_of(int)

    @staticmethod
    def _tag_sequence(
        annotations: Collection[AnnotatedSpan]
    ) -> Iterable["HTMLStyleAnnotationFormatter.Tag"]:
        """
        Provide the tags in order of their occurrence in the formatted text.

        Each step yields a triple of a tag name, whether or not it is a start tag (a boolean),
        and the offset of the beginning of the tag.
        """

        # for start tags, we want to break ties by longest length so that tags which end later
        # start first. This preserves proper nesting.
        def start_tag_key(annotation):
            return annotation.span.start, -len(annotation.span)

        # for end tags, we want to break ties by shortest length so that tags which start later
        # end earlier. This preserves proper nesting.
        def end_tag_key(annotation):
            return annotation.span.end, len(annotation.span)

        start_tags = (
            HTMLStyleAnnotationFormatter.Tag(to_start_tag(ann), True, ann.span.start)
            for ann in sorted(annotations, key=start_tag_key)
        )
        end_tags = (
            HTMLStyleAnnotationFormatter.Tag(to_end_tag(ann), False, ann.span.end)
            for ann in sorted(annotations, key=end_tag_key)
        )

        # interleave the start and end tag lists
        # when start and end tags are attached to the same offset, put the ends tags first
        # to avoid crossing elements
        # "sorted" is stable so the relative order of start and end tags is maintained
        return sorted(
            itertools.chain(end_tags, start_tags),
            # True is less than False, which is fine because we want end tags
            # before start tags,
            key=lambda tag: (tag.offset, tag.is_start),
        )
